import asyncio
import signal
from typing import Callable

from fastapi import Request

# This is to allow the last request's response to finish handling. There may be more
# middlewares that the response goes through, and then there's the time for the bytes
# to be pushed to the caller.
DEFAULT_TERM_DELAY_SECS = 5.0


class TerminationHandlerMiddleware:
    """
    This middleware allows for swiftly and safely terminating the server. It
    listens to a set of termination signals. On receiving such a signal, it
    first informs on the on_stop callback, then waits for currently executing
    requests to finish, before informing on the on_term callback.

    Stop means that the process to stop the server has started. As soon as
    outstading requests go to zero after this, on_term will be called.

    Term means that this is the right time to terminate the server process, no
    outstanding requests at this point.

    The caller would typically handle on_stop by stop sending more requests to
    the FastApi server. And on_term by exiting the server process.
    """

    def __init__(
        self,
        on_stop: Callable[[], None],
        on_term: Callable[[], None],
        termination_delay_secs: float = DEFAULT_TERM_DELAY_SECS,
    ):
        self._outstanding_request_count = 0
        self._on_stop = on_stop
        self._on_term = on_term
        self._termination_delay_secs = termination_delay_secs
        self._stopped = False
        for sig in [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT]:
            signal.signal(sig, self._stop)

    async def __call__(self, request: Request, call_next):
        self._outstanding_request_count += 1
        try:
            response = await call_next(request)
        finally:
            self._outstanding_request_count -= 1
            if self._outstanding_request_count == 0 and self._stopped:
                # There's a delay in term to allow some time for current
                # response flow to finish.
                asyncio.create_task(self._term())
        return response

    def _stop(self, sig, frame):
        self._on_stop()
        self._stopped = True
        if self._outstanding_request_count == 0:
            self._on_term()

    async def _term(self):
        await asyncio.sleep(self._termination_delay_secs)
        self._on_term()



import os
from logging import Logger

import requests
from anyio import to_thread
from helpers.inference_server_controller import InferenceServerController
from tenacity import Retrying, stop_after_attempt, wait_exponential


def inference_server_startup_flow(
    inference_server_controller: InferenceServerController,
    logger: Logger,
) -> None:
    """
    Perform the inference server startup flow

    Inference server startup flow supports checking for patches. If a patch ping
    url is provided then we hit that url to start the sync mechanism. The ping
    calls with current truss hash. The patch ping endpoint should return a
    response indicating, either that the supplied hash is current or that the
    request has been accepted. Acceptance of request means that a patch will be
    supplied soon to the truss (by calling of /control/patch endpoint).

    If we find that our hash is current, we start the inference server
    immediately. Otherwise, we delay the start to when the patch is supplied.

    The goal is to start the inference server as soon as we have the latest
    code, but not before.
    Example responses:
    {"is_current": true}
    {"accepted": true}
    """
    patch_ping_url = os.environ.get("PATCH_PING_URL_TRUSS")
    if patch_ping_url is None:
        inference_server_controller.start()
        return

    truss_hash = inference_server_controller.truss_hash()
    payload = {"truss_hash": truss_hash}

    for attempt in Retrying(
        stop=stop_after_attempt(15),
        wait=wait_exponential(multiplier=2, min=1, max=4),
    ):
        with attempt:
            try:
                logger.info(
                    f"Pinging {patch_ping_url} for patch with hash {truss_hash}"
                )
                resp = requests.post(patch_ping_url, json=payload)
                resp.raise_for_status()
                resp_body = resp.json()

                # If hash is current start inference server, otherwise delay that
                # for when patch is applied.
                if "is_current" in resp_body and resp_body["is_current"] is True:
                    logger.info("Hash is current, starting inference server")
                    inference_server_controller.start()
            except Exception as exc:  # noqa
                logger.warning(f"Patch ping attempt failed with error {exc}")
                raise exc


async def async_inference_server_startup_flow(
    inference_server_controller: InferenceServerController,
    logger: Logger,
) -> None:
    return await to_thread.run_sync(
        inference_server_startup_flow, inference_server_controller, logger
    )



import enum
import logging
from typing import TYPE_CHECKING, Any, Dict, List

if TYPE_CHECKING:
    from python_on_whales.components.container.cli_wrapper import Container

from truss.constants import TRUSS_DIR
from truss.local.local_config_handler import LocalConfigHandler


class Docker:
    _client = None

    @staticmethod
    def client():
        if Docker._client is None:
            from python_on_whales import DockerClient, docker

            if LocalConfigHandler.get_config().use_sudo:
                Docker._client = DockerClient(client_call=["sudo", "docker"])
            else:
                Docker._client = docker
        return Docker._client


def get_containers(labels: Dict, all: bool = False):
    """Gets containers given labels."""
    return Docker.client().container.list(
        filters=_create_label_filters(labels), all=all
    )


def get_images(labels: Dict):
    """Gets images given labels."""
    return Docker.client().image.list(filters=_create_label_filters(labels))


def get_urls_from_container(container_details) -> Dict[int, List[str]]:
    """Gets url where docker container is hosted."""
    if (
        container_details.network_settings is None
        or container_details.network_settings.ports is None
    ):
        return {}
    ports = container_details.network_settings.ports

    def parse_port(port_protocol) -> int:
        return int(port_protocol.split("/")[0])

    def url_from_port_protocol_value(port_protocol_value: Dict[str, str]) -> str:
        return (
            "http://"
            + port_protocol_value["HostIp"]
            + ":"
            + port_protocol_value["HostPort"]
        )

    def urls_from_port_protocol_values(
        port_protocol_values: List[Dict[str, str]],
    ) -> List[str]:
        return [url_from_port_protocol_value(v) for v in port_protocol_values]

    return {
        parse_port(port_protocol): urls_from_port_protocol_values(value)
        for port_protocol, value in ports.items()
        if value is not None
    }


def kill_containers(labels: Dict[str, bool]) -> None:
    from python_on_whales.exceptions import DockerException

    containers = get_containers(labels)
    for container in containers:
        container_labels = container.config.labels
        if TRUSS_DIR in container_labels:
            truss_dir = container_labels[TRUSS_DIR]
            logging.info(f"Killing Container: {container.id} for {truss_dir}")
    try:
        Docker.client().container.kill(containers)
    except DockerException:
        # The container may have stopped running by this point, this path
        # is for catching that. Unfortunately, there's no separate exception
        # for this scenario, so we catch the general one. Specific exceptions
        # such as NoSuchContainer are still allowed to error out.
        pass


def get_container_logs(container, follow, stream):
    return Docker.client().container.logs(container, follow=follow, stream=stream)


class DockerStates(enum.Enum):
    CREATED = "created"
    RUNNING = "running"
    PAUSED = "paused"
    RESTARTING = "restarting"
    OOMKILLED = "oomkilled"
    DEAD = "dead"
    EXITED = "exited"


def inspect_container(container) -> "Container":
    """Inspects truss container"""
    return Docker.client().container.inspect(container)


def get_container_state(container) -> DockerStates:
    "Get state of the container"
    return DockerStates(inspect_container(container).state.status)


def _create_label_filters(labels: Dict) -> Dict[str, Any]:
    return {
        f"label={label_key}": label_value for label_key, label_value in labels.items()
    }



from typing import Dict, List

from packaging import requirements  # type: ignore


def identify_requirement_name(req: str) -> str:
    try:
        parsed_req = requirements.Requirement(req)
        return parsed_req.name  # type: ignore
    except (requirements.InvalidRequirement, ValueError):
        # default to the whole line if we can't parse it.
        return req.strip()


def reqs_by_name(reqs: List[str]) -> Dict[str, str]:
    return {identify_requirement_name(req): req for req in reqs}



from enum import Enum

import pydantic


class DeployedChainlet(pydantic.BaseModel):
    name: str
    is_entrypoint: bool
    status: str
    logs_url: str


class ChainletData(pydantic.BaseModel):
    name: str
    oracle_version_id: str
    is_entrypoint: bool


class ModelOrigin(Enum):
    BASETEN = "BASETEN"
    CHAINS = "CHAINS"
